# -*- coding: utf-8 -*-
"""
Created on Fri Jan 18 10:08:09 2019

@author: mcamp
"""
import os
from astropy.io import fits
from astropy.stats import gaussian_fwhm_to_sigma
from astropy.convolution import Gaussian2DKernel
from photutils import Background2D, MedianBackground
from photutils import detect_sources
from photutils import source_properties
from astropy import wcs
from astropy.visualization.mpl_normalize import ImageNormalize
from astropy.visualization import SqrtStretch, MinMaxInterval
from astropy import units as u
import numpy as np
import fnmatch
from astropy.table import Table


def load_wcs_from_file(filename):
    # Load the FITS hdulist using astropy.io.fits
    hdulist = fits.open(filename)

    # Parse the WCS keywords in the primary HDU
    w = wcs.WCS(hdulist[0].header)
    hdulist.close()

    # Three pixel coordinates of interest.
    # Note we've silently assumed a NAXIS=2 image here
    pixcrd = np.array([[0, 0], [24, 38], [45, 98]], np.float_)

    # Convert pixel coordinates to world coordinates
    # The second argument is "origin" -- in this case we're declaring we
    # have 1-based (Fortran-like) coordinates.
    world = w.wcs_pix2world(pixcrd, 1)

    # Convert the same coordinates back to pixel coordinates.
    pixcrd2 = w.wcs_world2pix(world, 1)

    # These should be the same as the original pixel coordinates, modulo
    # some floating-point error.
    assert np.max(np.abs(pixcrd - pixcrd2)) < 1e-6
    return w


def parse_wcs_from_hdu(hdulist):
    # Parse the WCS keywords in the primary HDU
    w = wcs.WCS(hdulist[0].header)

        # Three pixel coordinates of interest.
    # Note we've silently assumed a NAXIS=2 image here
    pixcrd = np.array([[0, 0], [24, 38], [45, 98]], np.float_)

    # Convert pixel coordinates to world coordinates
    # The second argument is "origin" -- in this case we're declaring we
    # have 1-based (Fortran-like) coordinates.
    world = w.wcs_pix2world(pixcrd, 1)
    # Convert the same coordinates back to pixel coordinates.
    pixcrd2 = w.wcs_world2pix(world, 1)

    # These should be the same as the original pixel coordinates, modulo
    # some floating-point error.
    valid_wcs = (np.max(np.abs(world - pixcrd)) > 1e-6) & (np.max(np.abs(pixcrd - pixcrd2)) < 1e-6)

    return w, valid_wcs


def create_sources_hdu(source_table):
    # Add sources to the original FITS file as a FITS binary table
    # Create columns
    x = fits.Column(name='x', array=source_table['xcentroid'], format='E')
    y = fits.Column(name='y', array=source_table['ycentroid'], format='E')
    eccentricity = fits.Column(name='eccentricity', array=source_table['eccentricity'], format='E')
    cols = fits.ColDefs([x, y, eccentricity])
    # Create a default header
    sources_hdr = fits.Header()
    # Add specific cards
    sources_hdr['TTYPE1'] = "X"
    sources_hdr['TTYPE2'] = "Y"
    sources_hdr['TTYPE3'] = "ECCENTRICITY"
    sources_hdr['EXTNAME'] = "SOURCES"
    sources_hdr['IMAGEW'] = image_data.shape[1]
    sources_hdr['IMAGEH'] = image_data.shape[0]
    # Add HDU and header to file
    sources_hdu = fits.BinTableHDU.from_columns(cols,  header=sources_hdr)
    return sources_hdu


def process_image(hdul):
    # Detect sources in a image

    # get thei mage data
    image_data = hdul[0].data
    try:
        # Try and lok for some existing detecions
        print('Looking for existing detections')
        sources_index = hdul.index_of('SOURCES')
    except:
        # None found so run image processing routine
        print('No pre-existing detections. Calculating sources')

        # Estimate background
        bkg_estimator = MedianBackground()
        bkg = Background2D(image_data, (50, 50), filter_size=(3, 3),bkg_estimator=bkg_estimator)
        # Calculate cutting threshold
        threshold = bkg.background + (2. * bkg.background_rms)
        # Create Gaussian kernel for convolution
        sigma = 3.0 * gaussian_fwhm_to_sigma    # FWHM = 3.
        kernel = Gaussian2DKernel(sigma, x_size=3, y_size=3)
        kernel.normalize()
        npixels = 5
        # Detect sources
        segm = detect_sources(image_data, threshold, npixels=npixels,filter_kernel=kernel)
        # Calculate the properties of the detected sources in the iamge
        cat = source_properties(image_data, segm)
        columns = ['xcentroid', 'ycentroid',  'eccentricity', 'area']
        source_table = cat.to_table(columns=columns)

        # Sort detected sources by size in descending order
        source_table.sort(['area'])
        source_table.reverse()
        # Remove any sources smaller than 30 pixels in area
        size_mask = (source_table['area'].to_value() > 30)
        source_table = source_table[size_mask]
        sources_hdu = create_sources_hdu(source_table)
        hdul.append(sources_hdu)
        # Create satellite and star tables

        satellite_mask = (source_table['eccentricity'].to_value()  > 0.95)
        # Satellite-like sources, i.e. streaks
        satellite_table = source_table[satellite_mask]
        # Star-like sources, i.e. circular sources
        star_table = source_table[~satellite_mask]
    else:
        # Found exisiting sources HDU
        print('Using pre-existing detections.')

        # Read the detections
        source_table = hdul[sources_index].data
        # Convert to table format from FITS binary table
        # Create satellite and star tables
        print(source_table.shape)
        satellite_mask = (source_table['eccentricity']  > 0.95)

        # Satellite-like sources, i.e. streaks
        satellite_table = source_table[satellite_mask]
        # Star-like sources, i.e. circular sources
        star_table = source_table[~satellite_mask]
    return satellite_table,star_table

# Directory where input files are
directory = r""

# Current file counter
file_index = 1
# Nubmer of files to process
number_of_files =  len(fnmatch.filter(os.listdir(directory), '*.fit')) + len(fnmatch.filter(os.listdir(directory), '*.fits'))
# loop over files in directory
for root, dirs, files in os.walk(directory):
    for file in files:
        if file.endswith('.fit') | file.endswith('.fits'):
            # Get full file path
            fits_image_filename = root+'\\'+file
            # Split file name and extensions
            file_identifier = os.path.splitext(os.path.basename(fits_image_filename))[0]
            file_extension = os.path.splitext(os.path.basename(fits_image_filename))[1]
            print("Processing = %d / %d" % (file_index, number_of_files))
            print(fits_image_filename)
            # Read in file and process
            with fits.open(fits_image_filename, mode='update') as hdul:
                # Get image data

                image_data = hdul[0].data
                hdul.info()
                print("File Read")
                # Parse the WCS frrom the header
                wcs_data,valid_wcs = parse_wcs_from_hdu(hdul)
                # Get the detected sources
                satellite_table,star_table = process_image(hdul)
                # Check to see if valid WCS, if not re-calibrate
                if valid_wcs == False:
                    raise AttributeError('Invalid WCS')
                sources_index = hdul.index_of('SOURCES')
                source_table = hdul[sources_index].data
                x = fits.Column(name='x', array=source_table['x'], format='E')
                y = fits.Column(name='y', array=source_table['y'], format='E')
                eccentricity = fits.Column(name='eccentricity', array=source_table['eccentricity'], format='E')
                ra,dec = wcs_data.all_pix2world(source_table['x'],source_table['y'], 0)
                ra_col = fits.Column(name='ra', array=ra, format='E')
                dec_col = fits.Column(name='dec', array=dec, format='E')
                cols = fits.ColDefs([x, y, ra_col, dec_col, eccentricity])
                sources_hdr = hdul[0].header.copy()
                sources_hdr['TTYPE1'] = "X"
                sources_hdr['TTYPE2'] = "Y"
                sources_hdr['TTYPE3'] = "RIGHT ASCENSION"
                sources_hdr['TTYPE4'] = "DECLINATION"
                sources_hdr['TTYPE5'] = "ECCENTRICITY"
                sources_hdr['EXTNAME'] = "SOURCES"
                sources_hdr['IMAGEW'] = image_data.shape[1]
                sources_hdr['IMAGEH'] = image_data.shape[0]
                # Add HDU and header to file
                try:
                    os.mkdir(root+"\\table")
                except OSError:
                    print ("Creation of the directory %s failed" % root+"\\table")
                else:
                    print ("Successfully created the directory %s " % root+"\\table")
                hdu = fits.BinTableHDU.from_columns(cols, header=sources_hdr)
                hdu.writeto(root+"\\table\\"+file_identifier+file_extension)

                file_index += 1
